/*
*	based on /sys/src/cmd/strings.c
*/
#include	<u.h>
#include 	<libc.h>
#include 	<ctype.h>
#include	<bio.h>
#include	"misc.h"
#define DBG if(0)

Biobuf	bout;

#define	MINSPAN		2		/* Minimum characters in string (default) */
#define BUFSIZE		70

void dodir(int fd, char *path);
void dofile(int fd);
int isokrune(Rune);

static int minlen = MINSPAN;
static dflag;
static hflag;
static iflag;
static aflag;
static oflag;
static nflag;


char *spec = " !\"#$%&'()*+,-./0123456789:;<=>?@[\\]^`{|}~";	// not used
/*
|20 sp |21  ! |22  " |23  # |24  $ |25  % |26  & |27  ' |
|28  ( |29  ) |2a  * |2b  + |2c  , |2d  - |2e  . |2f  / |
|30  0 |31  1 |32  2 |33  3 |34  4 |35  5 |36  6 |37  7 |
|38  8 |39  9 |3a  : |3b  ; |3c  < |3d  = |3e  > |3f  ? |
|40  @ 
|5b  [ |5c  \ |5d  ] |5e  ^ |5f  _ |
|60  ` 
|7b  { |7c  | |7d  } |7e  ~
*/



static void
usage(void)
{
	fprint(2, 
	"usage: %s [-l minlen] [-adhino] [path ...]\n"
	"  extwords extract words in files.\n"
	"  -d: the paths are directories\n"
	"  -l: minimum length of the words\n"
	"  -i: contents to lower case\n"
	"  -h: header name is last element of path\n"
	"  -hh: no header name\n"
	"  -a: add header info to contents (for indexing)\n"
	"  -o: add line offset\n"
	"  -n: numerics at the beginning are removed in extractiong words\n"
	, argv0);
	exits("usage");
}


/*	it seems rune 0x2000 to 0x2fff are special symbols */
#define xalpha(r) ((!(r & 0xff00) && isalpha(r)) || r == '_' || ((r&0xffff) > 0xa0 && ((r & 0xf000) != 0x2000)))
#define xalnum(r) ((!(r & 0xff00) && isalnum(r)) || r == '_' || ((r&0xffff) > 0xa0 && ((r & 0xf000) != 0x2000)))

/*	we would like include file name to index. because file
*	name is basic information.
*	we want to avoid register symbols to the index.
*	but what to do? consider examples below:
*	.123foo456  123foo456 .123foo456.tar	-> foo456
*	1234.gz					-> discard
*	dots := dot + ...
*	name := {dots + nodotname} + ...
*	nodotname := {spec + word} + ...
*	we use only first word	*/

void
extfn(char *fn)
/*	fn is scratched */
{
	char *s, *t;
	s = skip(fn,".");
	t = skipto(s, ".");
	*t = 0;
	while(*s && !xalpha(*s))
		s++;
	t = s;
	while(*t && xalnum(*t))
		t++;
	*t = 0;
	memcpy(fn, s, strlen(s) + 1);
}

char *
basename(char *path)
{
	char *p;
	p = strrchr(path,'/');
	if(p)
		p++;
	else
		p = path;
	return p;
}

void
main(int argc, char **argv)
{
	int i,n,fd;
	Dir *dir;
	char *p0,*p;

	ARGBEGIN{
	case 'd':
		dflag++;
		break;
	case 'l':
		minlen = atoi(EARGF(usage()));
		break;
	case 'h':
		hflag++;
		break;
	case 'a':
		aflag++;
		break;
	case 'i':
		iflag++;
		break;
	case 'o':
		oflag++;
		break;
	case 'n':
		nflag++;
		break;
	default:
		usage();
		break;
	}ARGEND

	Binit(&bout, 1, OWRITE);

	if(*argv == nil) {
		if(dflag){
			fd = open(".",OREAD);
			dodir(fd,".");
		}
		else
			dofile(0);
		exits(0);
	}



	if(dflag){
		for(i = 0; i < argc; i++) {
			/*	those argments may have trailing "/"
			*	if they have trailing "/", remove first	*/
			p = argv[i];
			n = strlen(p);
			if(p[n - 1] == '/')
				p[n - 1] = 0;
			fd = open(argv[i], OREAD);
			if(fd < 0)
				sysfatal("%s not open: %r",argv[i]);
			p0 = strdup(argv[i]);

			/*
			*	header output:
			*	case hflag
			*	0:  /dir_of_foo/foo:
			*	1:	foo:
			*	2:		# none
			*/

			if(!hflag)
				Bprint(&bout, "%s/:\n", p0);
			p = basename(p0);
			if(hflag == 1)
				Bprint(&bout, "%s/:\n", p);
			if(aflag){
				/*
				*	regulation to file name must be here
				*	it's late for mktoc to do this jos
				*/
				if(iflag)
					strtolower(p);
				extfn(p);
				/* we may discard the name less than minlen */
				if(utflen(p) >= minlen) 
					Bprint(&bout, "%s\n", p);
			}
			free(p0);
			dir = dirfstat(fd);
			if(dir->mode & DMDIR) /* directory */
				dodir(fd, argv[i]);
			close(fd);
		}
	}
	else{
		for(i = 0; i < argc; i++) {
			fd = open(argv[i], OREAD);
			if(fd < 0)
				sysfatal("%s not open: %r",argv[i]);
			p0 = strdup(argv[i]);
			if(!hflag)
				Bprint(&bout, "%s:\n", p0);
			p = basename(p0);
			if(hflag == 1)
				Bprint(&bout, "%s:\n", p);
			if(aflag){	
				if(iflag)
					strtolower(p);
				extfn(p);
				/*	how should I do if the file name is 1.c	?
				*	the extfn(p) will return "" for the name.
				* we may discard the name less than minlen */
				if(utflen(p) >= minlen) 
					Bprint(&bout, "%s\n", p);
			}
			free(p0);
			dir = dirfstat(fd);
			if((dir->mode & DMDIR) == 0) /* file */
				dofile(fd);
			free(dir);
			close(fd);
		}
	}
	exits(0);
}

#define RUNEERROR 0xFFFD

int
compar(Dir *a, Dir *b)
{
	return strcmp(a->name, b->name);
}


void
dodir(int fd, char *path)
{
	Dir *dir;
	char *s, *p;
	int i,n;
	int fd1;

	n = dirreadall(fd,&dir);
	if(n == -1)
		sysfatal("dirreadall: %r");
	if(n == 0){
		free(dir);
		return;
	}

	/* sort by name */
	qsort(dir, n, sizeof(Dir), (int (*)(const void*, const void*))compar);

	for(i = 0; i < n; i++)
		if((dir[i].mode & DMDIR) == 0){
			s = smprint("%s/%s",path,dir[i].name);
			fd1 = open(s, OREAD);
			if(fd1 < 0)
				sysfatal("%s not open: %r", s);
			if(!hflag)
				Bprint(&bout, "%s:\n", s);
			p = dir[i].name;
			if(hflag == 1)
				Bprint(&bout, "%s:\n", p);
			if(aflag){	
				if(iflag)
					strtolower(p);
				extfn(p);
				/* we may discard the name less than minlen */
				if(utflen(p) >= minlen) 
					Bprint(&bout, "%s\n", p);
			}
			free(s);
			dofile(fd1);
			close(fd1);
		}
	free(dir);
}

#ifdef XXX
Unicode table
http://www.unicode.org/charts/PDF/

0000:0080	C0 Controls and Basic Latin (US ASCII)
0080:0100	C1 Controls and Latin-1 Supplement
0100:0180	Latin Extended A
0180:0250	Latin Extended B
0250:02B0	IPA Extensions
02B0:0300	Spacing Modifier Letters
0300:0370	Combining Diacritical Marks
0370:0400	Greek and Coptic
0400:0500	Cyrillic
0500:0530	Cyrillic Supplement
0530:0590	Armenian
0590:0600	Hebrew
0600:0700	Arabic
0700:0750	Syriac
0750:0780	Arabic Supplement
1000:10A0	Myanmar
1100:1200	Hangul Jamo
1200:12C0	Ethiopic
1380:13A0	Ethiopic Supplement
1400:14E0	Unified Canadian Aboriginal Syllabics
2000:2070	General Punctuation
2070:20A0	Superscripts and Subscripts
20A0:20D0	Currency Symbols
20D0:2100	Combining Diacritical Marks for Symbols
2100:2150	Letterlike Symbols
2150:2190	Number Forms
2190:2200	Arrows
2200:2300	Mathematical Operators
2400:2440	Control Pictures
2440:2460	Optical Character Recognition
2460:2500	Enclosed Alphanumerics
2500:2580	Box Drawing
2580:25A0	Block Elements
2600:2700	Miscellaneous Symbols
2700:27C0	Dingbats
27C0:27F0	Miscellaneous Mathematical Symbols-A
27F0:2800	Supplemental Arrows-A
2800:2900	Braille Patterns
2900:2980	Supplemental Arrows-B
2980:2A00	Miscellaneous Mathematical Symbols-B
2A00:2B00	Supplemental Mathematical Operators
2B00:2C00	Miscellaneous Symbols and Arrows
2C60:2C80	Lattin Extended-C
2C80:2D00	Coptic
2D00:2D30	Georgian Supplement
2D30:2D80	Tifinagh
2D80:2DE0	Ethiopic Extended
2DE0:2E00	Cyrillic Extended-A
2E00:2E80	Supplemental Punctuation
2E80:2F00	CJK Radicals Supplement
2F00:2FE0	Kangxi Radicals
3040:30A0	Hiragana
30A0:3100	Katakana
3100:3130	Bopomofo
3130:3190	Hangul Compatibility Jamo
3190:31A0	Kanbun
3200:3300	Enclosed CJK Letters and Months
3300:3400	CJK Compatibility
3400:4DC0	CJK Unified Ideographs Extension A
4DC0:4E00	Yijing Hexagram Symbols
4E00:A000	CJK Unified Ideographs
A000:A490	Yi Syllables
A490:A4D0	Yi Radicals
A6A0:A700	Bamum
AC00:D800	hangul Syllables
F900:FAE0	CJK Compatibility Ideographs
FB00:FB50	Alphabetic Presentation Forms
FB50:FE00	Arabic Presentation Forms-A
FE00:FE10	Variation Selectors
FE30:FE4F	CJK Compatibility Forms
FE50:FE6F	Small Form Variants
FE70:FF00	Arabic Presentation Forms-B
FF00:FFF0	Halfwidth and Fullwidth Forms
FFF0:10000	Specials
#endif

int
zone(Rune r)
{
	if(r < 0x250)
		return 0x250;	// Latin
	if(r < 0x0370)
		return 0x0370; // etc
	if(r < 0x0400)
		return 0x0400; // Greek
	if(r < 0x0530)
		return 0x0530; // Cyrillic
	if(r < 0x3040)
		return 0x3040; // etc
	if(r < 0x30a0)
		return 0x30a0; // Hiragana
	if(r < 0x3100)
		return 0x3100; // Katakana
	if(r < 0x4e00)
		return 0x4e00; // etc
	if(r < 0xa000)
		return 0xa000; // CJK Kanji
	return 0x10000; // etc
}

/*	for binary check */
int
isokrune(Rune r)
{
	if (r == Runeerror)
		return 0;
	if(r == '\n' || r == '\t' || r == '\r')
		return 1;
	if(r > 0xA0)
		return 1;
	if (isprint(r))
		return 1;
	return 0;
}

/*	discard? or accept? */
int
acceptable(Rune r)
{
	int c = (r&0xffff);
	if(c < 0x80){
		if(isalnum(r) || r == '_')
			return 1;
		return 0;
	}
	/* I need help
	* I am not sure: accents or punctuations or symbols */

	/*	0080:00c0 C1 controls and symbols
	*	0xd7 is ×
	*	0xf7 is ÷ 	*/
	if(c < 0xc0 || c == 0xd7 || c == 0xf7)
		return 0;

	/*
	*	00c0:0100 Latin-1 supplement
	*	0100:0180 Latin extended A
	*	0180:0250 Latin extended B
	*	0250:02b0 IPA extensions
	*	02b0:0300 spacing modifier
	*	0300:0370 combining diacritical mark
	*	0370:0400 Greek and Coptic
	*	0400:0500 Cyrillic
	*	0500:0530 Cyrillic supplement
	*	0530:0590 Armenian (including symbols)
	*	the details: 0559:0560 and 0589:0590 are symbols
	*/
	if(c < 0x0370) // Latin regular letters + accent
		return 1;
	if(c < 0x0559) // Greek, Coptic, Cyrillic, and some of Armenian
		return 1;
	if(c < 0x0560) // Armenian symbols
		return 0;
	if(c < 0x0589) // Armenian
		return 1;
	if(c < 0x0590) // Armenian symbols
		return 0;

	/*
	*	0590:0600 Hebrew
	*	0600:0700 Arabic
	*	0700:0750 Syriac
	*	...
	*	I donn't know these language. for the temporally, all accept.
	*/
	if(c < 0x2000) // I don't know arabic etc. for the temporally, all accept.
		return 1;

	if(c < 0x2e80) // synbols such as math etc.
		return 0;
	if(c < 0x2fe0) // some CJK kanji
		return 1;
	if(c < 0x30a0) // some kanji braces etc and hiragana
		return 0;
	if(c < 0x3200) // katakana and others
		return 1;
	if(c < 0x3400) // CJK compatibility symbols
		return 0;
	if(c < 0x4dc0) // CJK Unified Ideographs Extension
		return 1;
	if(c < 0x4e00) // Yijing Hexagram Symbols
		return 0;
	if(c < 0xa000) // CJK Unified Ideographs
		return 1;
	if(c < 0xa4d0) // Yi Syllables and Radicals
		return 0;
	/* I don't know others. */
	return 1;
}

#define MAX 100

void
dofile(int fd)
{
	Biobuf bin;
	int cnt = 0;
	int c, c0;
	vlong offset;// line offset
	Rune buf[BUFSIZE];
	int nrune;
	int nerr;

	Binit(&bin, fd, OREAD);
	if(oflag){
		offset = Boffset(&bin);
		Bprint(&bout, "* %lld\n", offset);
	}

	c0 = nrune = nerr = 0;
	while((c = Bgetrune(&bin)) >= 0) {
		nrune++;
		if(!isokrune(c)){
			nerr++;
			if(nrune > MAX && 100 * nerr / nrune > 5) /* allow 5% err */
				break;
			continue;
		}
#ifdef XXX
		/*	for the temporally we should remove this one
		*	because this does not collaborate with grep	*/
		/*	FF00:FF60	Halfwidth and Fullwidth Forms
		*	convert to regular ascii	*/
		if(c >= 0xff00 && c < 0xfff0)
			c = c - 0xff00 + 0x20;
#endif
		if(acceptable(c)){
			if(nflag && cnt == 0 && isdigit(c))
				continue;
			if(iflag)
				c = tolowerrune(c);
				//c = tolower(c);// old version
			if(zone(c0) != zone(c)){
				buf[cnt] = 0;
				if(cnt >= minlen){
					if(oflag)
						Bprint(&bout, "%S %lld\n", buf, offset);
					else
						Bprint(&bout, "%S\n", buf);
				}
				cnt = 0;
				if(nflag && isdigit(c))
					continue;
			}
			buf[cnt++] = c;
			if(cnt == BUFSIZE-1) {
				buf[cnt] = 0;
				if(oflag)
					Bprint(&bout, "%S... %lld\n", buf, offset);
				else
					Bprint(&bout, "%S...\n", buf);
				cnt = 0;
			}
			c0 = c;
		}
		else{ /* space, symbol, etc */
			if(cnt >= minlen) {
				buf[cnt] = 0;
				if(oflag)
					Bprint(&bout, "%S %lld\n", buf, offset);
				else
					Bprint(&bout, "%S\n", buf);
			}
			cnt = 0;
			if(c == '\n' && oflag){
				offset = Boffset(&bin);
				Bprint(&bout, "* %lld\n", offset);
			}
		}
	}
	if(cnt >= minlen) {
		buf[cnt] = 0;
		if(oflag)
			Bprint(&bout, "%S %lld\n", buf, offset);
		else
			Bprint(&bout, "%S\n", buf);
	}
	if(oflag)
		Bprint(&bout, "* %lld\n", Boffset(&bin)); // for the file without '\n' at EOF
	Bflush(&bout);
	Bterm(&bin);
}
