/*
*	grep -f
*	is buggy, so I wrote ffm
*	ffm: file to field maching
*	ffm is useful, because grep does not have concept for field
*	thus, it is hard or impossible for grep to handle some types of problems.
*	-Kenar-
*/

#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <bio.h>
#include "trans.h"
#include "misc.h"

#define DBG if(0)
#define	DBG0	if(0)
#define MAXFIELD 10

void* (*Hget)(char*);

Biobuf bout;
char *sep;
int fnum;
int vflag;
int mflag;
int Kflag;
int forceflag;

typedef struct Hkeys Hkeys;
struct Hkeys{
	char *key;
	void *value;
	Hkeys *next;
} **hkeys;

/*	nhkey should be prime number
*	use command primes to find primes
*	primes 250 270
*	251
*	257
*	263
*	269		*/
int nhkeys =251;
int keylen;
#define MAPSIZE 1024
int *mpat;
int hkmin;
char *hkmap;
int npat;

/*	a simple linear hash */
int
mhash(char *s, int n)
{
	int h = 0;
	while(n-- > 0) h = (h << 1)%nhkeys + *s++;
	h %= nhkeys;
	return h<0?h+nhkeys:h;
}

int
hash(char *s)
{
	return mhash(s, strlen(s));
}


void
Hput(char *key, void *value)
{
	int h = hash(key);
	Hkeys *p = new(Hkeys);
	p->key = key;
	p->value = value;
	p->next = hkeys[h];
	hkeys[h] = p;
}

void *
Hget0(char *key)
/*	exact matching	*/
{
	Hkeys *p;
	for(p = hkeys[hash(key)];p;p = p->next){
		if(strcmp(p->key, key)==0)
			return p->value;
	}
	return nil;
}

void *
Hget1(char *key)
/*	partial matching	*/
{
	int k;
	Hkeys *p;

	if(!hkmap[mhash(key,hkmin)])
		return nil;

	for(k = npat - 1; k >= 0; k--){
		if(strlen(key) < mpat[k])
			continue;
		for(p = hkeys[mhash(key,mpat[k])];p;p = p->next){
			if(strlen(p->key) != mpat[k])
				continue;
			if(strncmp(p->key, key, strlen(p->key))==0)
				return p->key;
		}
	}
	return nil;
}


/*
*	patread: read data from pattern file	*/
void
patread(char *path)
{
	int k,n,m;
	char *p;
	List *q;
	List *lroot;
	char *map;
	int kmax,kmin;
	n = filetolist(&lroot, path);
	if(n == 0){
		forceflag = 1;
		return;
	}
	nhkeys = upprime(n);
	DBG fprint(2,"nhkeys=%d (%d)\n",nhkeys,n);
	hkeys = (Hkeys**) calloc(nhkeys, sizeof(Hkeys*));
	n = 0; kmax = 0; kmin = 9999999;
	for(q = lroot; q; q = q->next){
		p = q->data;
		if(*p == 0)
			sysfatal("empty line in %s",path);
		k = strlen(p);
		if(k > kmax)
			kmax = k;
		if(k < kmin)
			kmin = k;
		Hput(p,"");
		n++;
	}

	if(!mflag)
		return;

	hkmap = calloc(nhkeys, sizeof(char));// hash key map
	for(q = lroot; q; q = q->next)
		hkmap[mhash(q->data,kmin)] = 1;
	hkmin = kmin;

	map = calloc(kmax+1, sizeof(char));

	for(q = lroot; q; q = q->next){
		p = q->data;
		map[strlen(p)] = 1;
	}
	m = 0;
	for(k = 1; k <= kmax; k++){
		if(map[k])
			m++;
	}
	mpat = calloc(m+1, sizeof(int));
	m = 0;
	for(k = 1; k <= kmax; k++){
		if(map[k]){
			mpat[m] = k;
			m++;
		}
	}
	mpat[m] = 0;
	npat = m;
}


void
doit(int fd)
{
	char *p, *s, *args[MAXFIELD];
	char *buf; /* pointer to buffer */
	Rdbuf *rb;
	int m,n;
	int bufsize = 1024; /* initial buf size */
	rb = rdinit(fd, 32*1024);

	if(forceflag){
		if(vflag){
			while((p = rdline(rb))) /* assign = */
				Bprint(&bout, "%s\n", p);
			Bflush(&bout);
		}
		return;
	}

	buf = emalloc(bufsize);

	while((p = rdline(rb))){ /* assign = */
		if(rdlnlen(rb) > bufsize - 1){
			free(buf);
			bufsize = rdlnlen(rb) + 1024; /* at least + 1 */
			buf = emalloc(bufsize);
		}
		/*	strncpy() is very slow.
		*	that does needless job. use strecpy instead	*/
		strecpy(buf, buf + bufsize, p);

		/*	Why getfields() is slower than tokenize() ?
		*	this is because getfields() allows utf coded chars
		*	in the separater. it would be much simpler one
		*	if the separater is a single char and ascii	*/
		/*	note on split()
		*	in special case, if *sep == 0 then split() returns 1
		*	and args[0] is the line in buf */
		if(sep)
			m = split(buf, args, MAXFIELD, *sep);
		else
			m = tokenize(buf, args, MAXFIELD);
		if(m < fnum)
			continue;
		if(fnum == 0){
			for(n = 0; n < m; n++){
				s = Hget(args[n]);
				if((!vflag && s) || (vflag && !s)){
					if(Kflag && s)
						Bprint(&bout, "%s%s%d\n", p,sep,strlen(s));
					else
						Bprint(&bout, "%s\n", p);
				}
			}
		}
		else{
			s = Hget(args[fnum-1]);
			if((!vflag && s) || (vflag && !s)){
				if(Kflag && s)
					Bprint(&bout, "%s%s%d\n", p,sep,strlen(s));
				else
					Bprint(&bout, "%s\n", p);
			}
		}
	}
	free(buf);
	Bflush(&bout);
}

#ifdef XXX
/*
*	this isn't used, because it is too slow.
*	but I keep it so that I don't write it again.
*/
void
doit(int fd)
{
	Biobufhdr bin;
	char *p, *s, *args[MAXFIELD];
	char buf[MAXBUF];
	uchar biobuf[32*1024];
	int m,n;

	if(Binits(&bin, fd, OREAD, biobuf, sizeof(biobuf)) < 0)
		sysfatal("input not open: %r");
	while((p = Brdstr(&bin,'\n',1))){ /* assign = */
		strncpy(buf,p,MAXBUF);
		if(sep)
			m = getfields(buf, args, MAXFIELD, 0, sep);
		else
			m = tokenize(buf, args, MAXFIELD);
		if(m < fnum)
			continue;
		if(fnum == 0){
			for(n = 0; n < m; n++){
				s = Hget(args[n]);
				if((!vflag && s) || (vflag && !s)) {
					Bwrite(&bout, p, strlen(p));
					Bwrite(&bout,"\n",1);
				}
			}
		}
		else{
			s = Hget(args[fnum-1]);
			if((!vflag && s) || (vflag && !s)){
				Bwrite(&bout, p, strlen(p));
				Bwrite(&bout,"\n",1);
			}
		}
	}
	Bflush(&bout);
}
#endif

void
usage(void)
{
	fprint(2,
	"usage: ffm [-f $num] [-F $sep] [-v] [-m] [-K] $pat_file [$file ....]\n"
	"	file to field maching.\n"
	"	show lines if a field matches to one of strings in $pat_file.\n"
	"	without option \"-m\", the match is exact matching.\n"
	" -f $num	# field number that begins with 1,2,...\n"
	" -F $sep	# field separator\n"
	" -v		# show lines that do not match the pattern.\n"
	" -m		# substring match from the beginning of field.\n"
	" -K		# append matched key length at the end of field\n"
	"if $num not be given, then fmp shows lines if one of the fields matches\n"
	"to one of strings in $pat_file\n"
	);
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int fd;
	Hget = Hget0;
	ARGBEGIN{
	case 'F': sep = EARGF(usage());
		break;
	case 'K': Kflag++;
		break;
	case 'f': fnum = atoi(EARGF(usage()));
		break;
	case 'm': mflag++;
		break;
	case 'v': vflag++;
		break;
	default: usage();
	}ARGEND

	if(*argv == nil)
		usage();

	Binit(&bout, 1, OWRITE);

	patread(*argv++);

	if(mflag)
		Hget = Hget1;

	if(*argv == nil){
		doit(0);
		exits(nil);
	}

	while(*argv){
		fd = open(*argv, OREAD);
		if(fd < 0)
			sysfatal("%s not open: %r",*argv);
		doit(fd);
		argv++;
		close(fd);
	}
	exits(nil);
}
